/***************************************************************************
 *   Copyright (C) 2014 by Abderrahman Taha                                *
 *                                                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor,Boston, MA 02110-1301 USA             *
 ***************************************************************************/

#include <QApplication>
#include "ui_modules/drawingoptions.h"
#include "ui_modules/colorsoptions.h"
#include "ui_modules/about.h"
#include "gl_viewer/glviewer.h"
#include "ui_modules/mathmod.h"
#include "json_parser/parisodef.h"

#ifdef Q_WS_X11
#include <X11/Xlib.h>
#endif

int main(int argc, char *argv[])
{
    QApplication app(argc, argv);

    //Creation of the two most important object:
    MathMod mm(0, "Iso");
    DrawingOptions drawingopt;

    //save references to each other to let theses object communicate:
    drawingopt.MathmodRef = &mm;
    mm.Parent = &drawingopt;


    drawingopt.MathmodRef->RootObjet.ReadJsonFile("./mathmodcollection.js", drawingopt.JSONMathModels);

    QJsonArray array = drawingopt.JSONMathModels["MathModels"].toArray();
    QStringList lst;
    QString a;
    QJsonObject jsobj;

    for(int i=0; i< array.size(); i++)
    {
        if((array[i].toObject())["Iso3D"].isObject())
        {
            jiso newjiso;
            jsobj = (array[i].toObject())["Iso3D"].toObject();
            newjiso.read( jsobj );
            drawingopt.MathmodRef->pariso.JIso.append(newjiso);
            a = (jsobj)["Name"].toArray()[0].toString();
            lst.append(a);
        }
        else if((array[i].toObject())["Param3D"].isObject())
        {
            jpar newjpar;
            jsobj = ((array[i].toObject())["Param3D"].toObject());
            newjpar.read( jsobj );
            drawingopt.MathmodRef->pariso.JPar.append(newjpar);
            a = ( jsobj )["Name"].toArray()[0].toString();
            lst.append(a);
        }
        else if((array[i].toObject())["Param4D"].isObject())
        {
            jpar newjpar;
            jsobj = ((array[i].toObject())["Param4D"].toObject());
            newjpar.read( jsobj );
            drawingopt.MathmodRef->pariso.JPar.append(newjpar);
            a = ( jsobj )["Name"].toArray()[0].toString();
            lst.append(a);
        }
    }

    lst.insert(0, "Examples ("+QString::number(lst.count())+")");
    //Load the script containing isosurface and parametric formulas:
    drawingopt.ui.choice->insertItems(0, lst);
    drawingopt.AddListModels();

    //Show the two windows of the application:
    drawingopt.move(10,10);
    mm.move(800,10);
    mm.show();
    drawingopt.show();

    return app.exec();
}
