/***************************************************************************
 *   Copyright (C) 2014 by Abderrahman Taha                                *
 *                                                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor,Boston, MA 02110-1301 USA             *
 ***************************************************************************/
#include "Model3D.h"

//+++++++++++++++++++++++++++++++++++++++++
Par3D::~Par3D()
{
}

//+++++++++++++++++++++++++++++++++++++++++
Par3D::Par3D()
{
    initialiser_parametres();
    initialiser_parseur();
}

//+++++++++++++++++++++++++++++++++++++++++
void Par3D::initialiser_parametres()
{
    nb_licol = 25;
    nb_ligne = nb_colone = nb_licol;
    largeur_fenetre = 620;
    hauteur_fenetre = 620;
    mesh = 1;
    infos =1;
    latence = 30;
    stepMorph = 0;
    pace = (double)1/(double)30;
    coupure_col = coupure_ligne = 0;
    activeMorph = -1;
    tetazw = tetaxy =  tetaxz = tetayz = tetaxw = tetayw =  0;
    tetazw_ok = tetaxy_ok =  tetaxz_ok = tetayz_ok = tetaxw_ok = tetayw_ok =  param4D = -1;
    // initialisation des matrices 4D
    mat4D                                     = Matrix4D();
    mat_rotation4D                      = Matrix4D();
    mat_rotation_save4D           = Matrix4D();
    mat_homothetie4D               = Matrix4D();
    mat_translation4D                 = Matrix4D();
    mat_inversetranslation4D    = Matrix4D();
    mat4D.unit();
}

//+++++++++++++++++++++++++++++++++++++++++
void Par3D::initialiser_parseur()
{
    expression_X = "(1+ 0.5*cos(t*u))*cos(v) -2";
    expression_Y = "(1+ 0.5*cos(t*u))*sin(v) -2";
    expression_Z = "0.5*sin(u) -2";
    expression_CND = "x > 0";
    inf_u = "0";
    inf_v = "0";
    sup_u = "2*pi";
    sup_v = "2*pi";
    myParser.AddConstant   ("pi", 3.14159265);
    myParserX.AddConstant("pi", 3.14159265);
    myParserY.AddConstant("pi", 3.14159265);
    myParserZ.AddConstant("pi", 3.14159265);
    myParserCND.AddConstant("pi", 3.14159265);
}

//+++++++++++++++++++++++++++++++++++++++++
void  Par3D::parse_expression()
{
    double vals[] = { 0,0,0};

    myParser.Parse(inf_u, "u,v,t");
    u_inf = myParser.Eval(vals);
    myParser.Parse(sup_u, "u,v,t");
    u_sup = myParser.Eval(vals);
    dif_u = u_sup - u_inf;

    myParser.Parse(inf_v, "u,v,t");
    v_inf = myParser.Eval(vals);
    myParser.Parse(sup_v, "u,v,t");
    v_sup = myParser.Eval(vals);
    dif_v = v_sup - v_inf;

    myParserX.Parse(expression_X, "u,v,t");
    myParserY.Parse(expression_Y, "u,v,t");
    myParserZ.Parse(expression_Z, "u,v,t");
    if(param4D == 1)
        myParserW.Parse(expression_W, "u,v,t");
}

//++++++++++++++++++++++++++++++++++++++++
void  Par3D::rotation4()
{
    mat_rotation4D.unit();
    // Construction de la matrice de trnsformation
    if(tetaxy_ok == 1)    mat_rotation4D.xyrot(tetaxy);
    if(tetaxz_ok == 1)    mat_rotation4D.xzrot(tetaxz);
    if(tetayz_ok == 1)    mat_rotation4D.yzrot(tetayz);

    if(tetaxw_ok == 1)    mat_rotation4D.xwrot(tetaxw);
    if(tetayw_ok == 1)    mat_rotation4D.ywrot(tetayw);
    if(tetazw_ok == 1)    mat_rotation4D.zwrot(tetazw);

// On applique cette transformation a la matrice principale "mat"
    mat4D.mult(mat_rotation4D);
}

//+++++++++++++++++++++++++++++++++++++++++
void  Par3D::boite_englobante4D(int idx)
{
    MINX =999999999;
    MINY =999999999;
    MINZ =999999999;
    MINW =999999999;

    MAXX =-999999999;
    MAXY =-999999999;
    MAXZ =-999999999;
    MAXW =-999999999;

    int IDX = 0;
    for (int i=0; i < nb_ligne; i++)
        for (int j=0; j < nb_colone; j++)
        {
            if(MINX > NormVertexTab[IDX + 3 + idx*6] ) MINX = NormVertexTab[IDX + 3 + idx*6];
            if(MINY > NormVertexTab[IDX + 4 + idx*6] ) MINY = NormVertexTab[IDX + 4 + idx*6];
            if(MINZ > NormVertexTab[IDX + 5 + idx*6] ) MINZ = NormVertexTab[IDX + 5 + idx*6];
            if(MINW > ExtraDimension[i*nb_ligne + j + idx] ) MINW = ExtraDimension[i*nb_ligne + j + idx];

            if(MAXX < NormVertexTab[IDX + 3 + idx*6] ) MAXX = NormVertexTab[IDX + 3 + idx*6];
            if(MAXY < NormVertexTab[IDX + 4 + idx*6] ) MAXY = NormVertexTab[IDX + 4 + idx*6];
            if(MAXZ < NormVertexTab[IDX + 5 + idx*6] ) MAXZ = NormVertexTab[IDX + 5 + idx*6];
            if(MAXW < ExtraDimension[i*nb_ligne + j + idx] ) MAXW = ExtraDimension[i*nb_ligne + j + idx];
            IDX +=6;
        }

    DIFX = MAXX - MINX ;
    DIFY = MAXY - MINY ;
    DIFZ = MAXZ - MINZ ;
    DIFW = MAXW - MINW ;
// Recherche du maximum :
    DIFMAXIMUM = DIFX;
    if (DIFY > DIFMAXIMUM)
    {
        DIFMAXIMUM = DIFY;
    };
    if (DIFZ > DIFMAXIMUM)
    {
        DIFMAXIMUM = DIFZ;
    };
    if (DIFW > DIFMAXIMUM)
    {
        DIFMAXIMUM = DIFW;
    };
// On va inclure cet objet dans un HperCube de langueur maximum
// egale a "hauteur_fenetre"

    double decalage_xo  = -(MINX +MAXX)/2 ;
    double decalage_yo  = -(MINY +MAXY)/2 ;
    double decalage_zo  = -(MINZ +MAXZ)/2 ;
    double decalage_wo = -(MINW +MAXW)/2 ;
    IDX =0;
    for (int i=0; i < nb_ligne   ; i++)
        for (int j=0; j < nb_colone   ; j++)
        {
            NormVertexTab[IDX + 3 + idx*6]= (NormVertexTab[IDX + 3 + idx*6] + decalage_xo)/DIFMAXIMUM ;
            NormVertexTab[IDX + 4 + idx*6] = (NormVertexTab[IDX + 4 + idx*6] + decalage_yo)/DIFMAXIMUM ;
            NormVertexTab[IDX + 5 + idx*6] = (NormVertexTab[IDX + 5 + idx*6] + decalage_zo)/DIFMAXIMUM ;
            ExtraDimension[i*nb_ligne + j + idx] = (ExtraDimension[i*nb_ligne + j + idx] + decalage_wo)/DIFMAXIMUM ;
            IDX +=6;
        }
}

//+++++++++++++++++++++++++++++++++++++++++
void  Par3D::Invert_boite_englobante4D(int idx)
{
    double decalage_xo  = -(MINX +MAXX)/2;
    double decalage_yo  = -(MINY +MAXY)/2;
    double decalage_zo  = -(MINZ +MAXZ)/2;
    int IDX =0;
    for (int i=0; i < nb_ligne   ; i++)
        for (int j=0; j < nb_colone   ; j++)
        {
            NormVertexTab[IDX + 3 + idx*6] = (NormVertexTab[IDX + 3 + idx*6]*DIFMAXIMUM -  decalage_xo);
            NormVertexTab[IDX + 4 + idx*6] = (NormVertexTab[IDX + 4 + idx*6]*DIFMAXIMUM -  decalage_yo);
            NormVertexTab[IDX + 5 + idx*6] = (NormVertexTab[IDX + 5 + idx*6]*DIFMAXIMUM -  decalage_zo);
            IDX +=6;
        }
}

//+++++++++++++++++++++++++++++++++++++++++
void Par3D::Anim_Rot4D (int idx)
{
    rotation4();
    calcul_points4(idx);         // On applique la rotation 4D
    boite_englobante4D(idx);
    project_4D_to_3D(idx);
    Invert_boite_englobante4D(idx);
}

//+++++++++++++++++++++++++++++++++++++++++
void  Par3D::calcul_points4(int idx)
{
    int i,j;
    double tp1, tp2, tp3, tp4;
// Changement de coordonnees des points selon les
// angles angex et angley
    int lndex =0;
    for (i=0; i < nb_ligne  ; i++)
        for (j=0; j < nb_colone   ; j++)
        {
            tp1 = NormVertexTab[lndex + 3 + idx*6];
            tp2 = NormVertexTab[lndex + 4 + idx*6];
            tp3 = NormVertexTab[lndex + 5 + idx*6];
            tp4 = ExtraDimension[i*nb_ligne + j + idx];

            NormVertexTab[lndex + 3 + idx*6] = mat4D.xx*tp1 + mat4D.xy*tp2 + mat4D.xz*tp3 + mat4D.xw*tp4 + mat4D.xo;
            NormVertexTab[lndex + 4 + idx*6] = mat4D.yx*tp1 + mat4D.yy*tp2 + mat4D.yz*tp3 + mat4D.yw*tp4 + mat4D.yo;
            NormVertexTab[lndex + 5 + idx*6] = mat4D.zx*tp1 + mat4D.zy*tp2 + mat4D.zz*tp3 + mat4D.zw*tp4 + mat4D.zo;
            ExtraDimension[i*nb_ligne + j + idx] = mat4D.wx*tp1 + mat4D.wy*tp2 + mat4D.wz*tp3 + mat4D.ww*tp4 + mat4D.wo;
            lndex += 6;
        }
}

//+++++++++++++++++++++++++++++++++++++++++++
void  Par3D::project_4D_to_3D(int idx)
{
    double c4;
    int I = 0;
    for (int i=0; i < nb_ligne; ++i)
        for (int j=0; j < nb_colone  ; ++j)
        {
            c4 = 1/(ExtraDimension[i*nb_ligne + j + idx] - 2);
            NormVertexTab[I + 3 + idx*6] *= c4;
            NormVertexTab[I + 4 + idx*6] *= c4;
            NormVertexTab[I + 5 + idx*6] *= c4;
            I += 6;
        }
}

//+++++++++++++++++++++++++++++++++++++++++++
void  Par3D::calcul_objet(int NewPosition, /*bool morph,*/ int cmp)
{
    double vals[] = {0,0,0};
    double iprime, jprime;

    if(stepMorph < 0)
    {
        pace = (double)1/(double)30;
    }
    if(stepMorph > 1)
    {
        pace = -(double)1/(double)30;
    }
    if((cmp == 0) && (activeMorph == 1))
        stepMorph += pace;

    vals[2]          = stepMorph;
    int l=0;
    for(int j=0; j < nb_ligne   ; j++)
    {
        jprime = (double)j/(double)(nb_ligne -1 ) ;
        jprime = jprime * dif_u  + u_inf;
        for (int i=0; i < nb_colone   ; i++)
        {
            iprime = (double)i/(double)(nb_colone-1 ) ;
            iprime = iprime * dif_v + v_inf;
            vals[0]=jprime;
            vals[1]=iprime;
            NormVertexTab[l+3+NewPosition] = myParserX.Eval(vals);
            NormVertexTab[l+4+NewPosition] = myParserY.Eval(vals);
            NormVertexTab[l+5+NewPosition] = myParserZ.Eval(vals);
            l+=6;
            if(param4D == 1)
                ExtraDimension[j*nb_colone + i + (int)(NewPosition/6)] = myParserW.Eval(vals);
        }
    }
}

//+++++++++++++++++++++++++++++++++++++++++
int Par3D::HowManyParamSurface(std::string ParamFct, int type)
{
    std::string tmp, tmp2;
    int position =0, jpos;
    int Nb_paramfunction =0;

    switch(type)
    {
    case 0:
        ParamStructs[0].fx = ParamFct;
        break;
    case 1:
        ParamStructs[0].fy = ParamFct;
        break;
    case 2:
        ParamStructs[0].fz = ParamFct;
        break;
    case 3:
        ParamStructs[0].umin = ParamFct;
        break;
    case 4:
        ParamStructs[0].umax = ParamFct;
        break;
    case 5:
        ParamStructs[0].vmin = ParamFct;
        break;
    case 6:
        ParamStructs[0].vmax = ParamFct;
        break;
    case 7:
        ParamStructs[0].fw = ParamFct;
        break;
    case 8:
        ParamStructs[0].cnd = ParamFct;
        break;
    }

    while( ParamFct!= "")
    {
        if((position = ParamFct.find(";")) >0   )
        {
            tmp = ParamFct;
            switch(type)
            {
            case 0:
                ParamStructs[Nb_paramfunction].fx = (tmp.substr(0,position));
                break;
            case 1:
                ParamStructs[Nb_paramfunction].fy = (tmp.substr(0,position));
                break;
            case 2:
                ParamStructs[Nb_paramfunction].fz = (tmp.substr(0,position));
                break;
            case 7:
                ParamStructs[Nb_paramfunction].fw = (tmp.substr(0,position));
                break;
            case 8:
                ParamStructs[Nb_paramfunction].cnd = (tmp.substr(0,position));
                break;
            case 3:
                ParamStructs[Nb_paramfunction].umin = (tmp.substr(0,position));
                break;
            case 4:
                ParamStructs[Nb_paramfunction].umax = (tmp.substr(0,position));
                break;
            case 5:
                ParamStructs[Nb_paramfunction].vmin = (tmp.substr(0,position));
                break;
            case 6:
                ParamStructs[Nb_paramfunction].vmax = (tmp.substr(0,position));
                break;
            }

            Nb_paramfunction++;
            tmp2 = ParamFct.substr(position+1, jpos=ParamFct.length()-1);
            ParamFct = tmp2;
        }
        else
        {
            switch(type)
            {
            case 0:
                ParamStructs[Nb_paramfunction].fx = ParamFct;
                break;
            case 1:
                ParamStructs[Nb_paramfunction].fy = ParamFct;
                break;
            case 2:
                ParamStructs[Nb_paramfunction].fz = ParamFct;
                break;
            case 7:
                ParamStructs[Nb_paramfunction].fw = ParamFct;
                break;
            case 8:
                ParamStructs[Nb_paramfunction].cnd = (tmp.substr(0,position));
                break;
            case 3:
                ParamStructs[Nb_paramfunction].umin = ParamFct;
                break;
            case 4:
                ParamStructs[Nb_paramfunction].umax = ParamFct;
                break;
            case 5:
                ParamStructs[Nb_paramfunction].vmin = ParamFct;
                break;
            case 6:
                ParamStructs[Nb_paramfunction].vmax = ParamFct;
                break;
            }
            ParamFct ="";
        }
    }
    return Nb_paramfunction;
}

//++++++++++++++++++++++++++++++++++++
void  Par3D:: ParamBuild(float *NormVertexTabPt,
                         unsigned int *IndexPolyTabPt,
                         unsigned int *PolyNumber,
                         unsigned int *VertxNumber,
                         int  IsoPos,
                         ComponentInfos *components
                         )
{
    std::string   expression_X_tmp   = expression_X,
                  expression_Y_tmp   = expression_Y,
                  expression_Z_tmp   = expression_Z,
                  expression_W_tmp  = expression_W,
                  inf_u_tmp = inf_u,
                  inf_v_tmp = inf_v,
                  sup_u_tmp = sup_u,
                  sup_v_tmp = sup_v;
    Nb_paramfunctions = HowManyParamSurface(expression_X, 0);
    HowManyParamSurface(expression_Y, 1);
    HowManyParamSurface(expression_Z, 2);
    HowManyParamSurface(inf_u, 3);
    HowManyParamSurface(sup_u, 4);
    HowManyParamSurface(inf_v, 5);
    HowManyParamSurface(sup_v, 6);
    if(param4D == 1)
        HowManyParamSurface(expression_W, 7);

    NbVertex  = (nb_ligne)*(nb_colone);
    NbPolygn = 4*(nb_ligne  - coupure_ligne -1)*(nb_colone - coupure_col -1);

    NormVertexTab = NormVertexTabPt;
    IndexPolyTab = IndexPolyTabPt;

    if(components != NULL)
        components->NbParametric = Nb_paramfunctions+1;
    int cutpoly, cut;
    cutpoly=cut=0;
    nbBorderPts = 0;
    // generate Isosurface for all the implicit formulas
    for(int fctnb= 0; fctnb< Nb_paramfunctions+1; fctnb++)
    {
        expression_X = ParamStructs[fctnb].fx;
        expression_Y = ParamStructs[fctnb].fy;
        expression_Z = ParamStructs[fctnb].fz;
        if(param4D == 1)
            expression_W = ParamStructs[fctnb].fw;
        expression_CND ="u > 0";
        inf_u = ParamStructs[fctnb].umin;
        sup_u = ParamStructs[fctnb].umax;
        inf_v = ParamStructs[fctnb].vmin;
        sup_v = ParamStructs[fctnb].vmax;

        parse_expression();
        calcul_objet(fctnb*6*NbVertex, fctnb);
        ///BorderCalculation(fctnb*6*NbVertex);

        if(param4D == 1)
        {
            Anim_Rot4D (fctnb*NbVertex);
        }

        calcul_Norm(fctnb*6*NbVertex);
        make_PolyIndex( fctnb,  IsoPos);

        /*
        cut = make_PolyIndex( fctnb,  IsoPos);
        cutpoly += cut;
        */

        if(components != NULL)
        {
            components->Parametricpositions[2*fctnb          ]  = fctnb*4*(nb_ligne  - coupure_ligne -1)*(nb_colone - coupure_col -1)  /*+  IsoPos*/; //save the starting position of this component
            components->Parametricpositions[2*fctnb     + 1] = (nb_ligne  - coupure_ligne -1)*(nb_colone - coupure_col -1) -(int)(cut/4); //save the number of Polygones of this component
        }
    }

    /*
    if(param4D == 1)
    {
        General_Anim_Rot4D (Nb_paramfunctions*NbVertex);
    }
    */

// 3) Nb Poly & Vertex :
    //*nbborderpoints = nbBorderPts;
    *PolyNumber      = (Nb_paramfunctions+1)*NbPolygn - cutpoly;
    *VertxNumber     = (Nb_paramfunctions+1)*NbVertex;

    expression_X = expression_X_tmp;
    expression_Y = expression_Y_tmp;
    expression_Z = expression_Z_tmp;
    if(param4D == 1)
        expression_W = expression_W_tmp;
    inf_u = inf_u_tmp;
    inf_v = inf_v_tmp;
    sup_u = sup_u_tmp;
    sup_v = sup_v_tmp;
}

//+++++++++++++++++++++++++++++++++++++++++++
void  Par3D::make_PolyIndex(int NewPo, int IsoPos)
{
    int k=0;
    int NewPosition = NewPo * 4*(nb_ligne  - coupure_ligne -1)*(nb_colone- coupure_col -1);
    int nbVertex       = NewPo * (nb_ligne  - coupure_ligne)*(nb_colone- coupure_col);
    for (int i=0; i< nb_ligne - coupure_ligne -1; i++)
        for (int j=0; j< nb_colone - coupure_col -1; j++)
        {
            IndexPolyTab[k    +NewPosition] =  i*nb_colone + j+nbVertex + IsoPos;
            IndexPolyTab[k+1+NewPosition] = (i+1)*nb_colone + j +nbVertex + IsoPos;
            IndexPolyTab[k+2+NewPosition] = (i+1)*nb_colone + (j+1)+nbVertex + IsoPos;
            IndexPolyTab[k+3+NewPosition] = i*nb_colone + (j+1)+nbVertex + IsoPos;
            k+=4;
        }
}

//+++++++++++++++++++++++++++++++++++++++++++
void  Par3D::calcul_Norm(int NewPosition)
{
//calculate Normals
    int        i, j, deplacement = 6*nb_colone;
    double caa, bab, cab, baa, ba, ca;

    for (i=0; i < nb_ligne -1  ; i++)
        for (j=0; j < nb_colone -1  ; j++)
        {
            caa = NormVertexTab[(i+1)*deplacement+j*6+4+NewPosition] - NormVertexTab[i*deplacement +j*6+4+NewPosition];
            bab = NormVertexTab[i*deplacement+(j+1)*6+5+NewPosition] - NormVertexTab[i*deplacement +j*6+5+NewPosition];
            cab = NormVertexTab[(i+1)*deplacement+j*6+5+NewPosition] - NormVertexTab[i*deplacement +j*6+5+NewPosition];
            baa = NormVertexTab[i*deplacement+(j+1)*6+4+NewPosition] - NormVertexTab[i*deplacement +j*6+4+NewPosition];
            ba   = NormVertexTab[i*deplacement+(j+1)*6+3+NewPosition] - NormVertexTab[i*deplacement +j*6+3+NewPosition];
            ca   = NormVertexTab[(i+1)*deplacement+j*6+3+NewPosition] - NormVertexTab[i*deplacement +j*6+3+NewPosition];

            NormVertexTab[i*deplacement +j*6     +NewPosition] = caa*bab  - cab*baa;
            NormVertexTab[i*deplacement +j*6 +1+NewPosition] = cab*ba    - ca*bab;
            NormVertexTab[i*deplacement +j*6 +2+NewPosition] = ca*baa    - caa*ba;

            double b4 = (double)sqrt(
                            (NormVertexTab[i*deplacement +j*6  +NewPosition]*NormVertexTab[i*deplacement +j*6      +NewPosition]) +
                            (NormVertexTab[i*deplacement +j*6+1+NewPosition]*NormVertexTab[i*deplacement +j*6+1+NewPosition]) +
                            (NormVertexTab[i*deplacement +j*6+2+NewPosition]*NormVertexTab[i*deplacement +j*6+2+NewPosition]));

            if( b4 < 0.00000001)  b4 = 0.00000001;

//Normalise:
            NormVertexTab[i*deplacement +j*6    +NewPosition]/=b4;
            NormVertexTab[i*deplacement +j*6+1+NewPosition]/=b4;
            NormVertexTab[i*deplacement +j*6+2+NewPosition]/=b4;
        }

    i = nb_ligne -1;
    for (j=0; j < nb_colone -1   ; j++)
    {
        NormVertexTab[i*deplacement +j*6    +NewPosition] = NormVertexTab[(i-1)*deplacement +j*6    +NewPosition];
        NormVertexTab[i*deplacement +j*6+1+NewPosition] = NormVertexTab[(i-1)*deplacement +j*6+1+NewPosition];
        NormVertexTab[i*deplacement +j*6+2+NewPosition] = NormVertexTab[(i-1)*deplacement +j*6+2+NewPosition];
    }

    j =nb_colone -1;
    for (i=0; i < nb_ligne -1  ; i++)
    {
        NormVertexTab[i*deplacement +j*6    +NewPosition] = NormVertexTab[i*deplacement +(j-1)*6    +NewPosition];
        NormVertexTab[i*deplacement +j*6+1+NewPosition] = NormVertexTab[i*deplacement +(j-1)*6+1+NewPosition];
        NormVertexTab[i*deplacement +j*6+2+NewPosition] = NormVertexTab[i*deplacement +(j-1)*6+2+NewPosition];
    }

    i = nb_ligne -1;
    j =nb_colone -1;
    NormVertexTab[i*deplacement +j*6  +NewPosition]    = NormVertexTab[(i-1)*deplacement +(j-1)*6    +NewPosition];
    NormVertexTab[i*deplacement +j*6+1+NewPosition]  = NormVertexTab[(i-1)*deplacement +(j-1)*6+1+NewPosition];
    NormVertexTab[i*deplacement +j*6+2+NewPosition]  = NormVertexTab[(i-1)*deplacement +(j-1)*6+2+NewPosition];
}
