/***************************************************************************
 *   Copyright (C) 2014 by Abderrahman Taha                                *
 *                                                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA            *
 ***************************************************************************/
#include <iostream>
#include <qstring.h>
#include <qfile.h>
#include <QTextStream>
#include <QStringList>

#include "../qjson/QJsonValue.h"
#include "../qjson/QJsonArray.h"
#include "../qjson/QJsonObject.h"
#include "../qjson/QJsonDocument.h"
#include "../qjson/QJsonParseError.h"
#include <QFileDialog>
#include <QDialog>
#include <QTranslator>

struct ParamStruct
{
    QString text;
    QString Varu;
    QString Const;
    QString Funct;
    QString Cnd;
    QString Grid;
    QString RGBT;
    QString fx;
    QString fy;
    QString fz;
    QString fw;
    QString umin;
    QString umax;
    QString vmin;
    QString vmax;
    QString tmin;
    QString tmax;
    QString comment;
    QString name;
    QString Component;
    int          Nbcomponent;
    int          index;
};

struct IsoStruct
{
    QString text;
    QString comment;
    QString name;
    QString Varu;
    QString Const;
    QString Funct;
    QString Cnd;
    QString Grid;
    QString RGBT;
    QString fxyz;
    QString xmin;
    QString xmax;
    QString ymin;
    QString ymax;
    QString zmin;
    QString zmax;
    QString Component;
    int          Nbcomponent;
    int          index;
};

struct TreeStruct
{
    QStringList name;
    QStringList Component;
    QStringList Varu;
    QStringList Const;
    QStringList Funct;
    QStringList Grid;
    QStringList Cnd;
    QStringList RGBT;
    QString       text;

    QStringList fxyz;
    QStringList xmin;
    QStringList xmax;
    QStringList ymin;
    QStringList ymax;
    QStringList zmin;
    QStringList zmax;

    QStringList fx;
    QStringList fy;
    QStringList fz;
    QStringList fw;
    QStringList umin;
    QStringList umax;
    QStringList vmin;
    QStringList vmax;
    QStringList tmin;
    QStringList tmax;
};

class ParisoMathObject
{
public :
    IsoStruct  IsoTable[1000];
    ParamStruct ParamTable[1000];

    int NbIsoStruct;
    int NbParamStruct;
    ParamStruct CurrentParamStruct;
    IsoStruct        CurrentIsoStruct;
    QJsonObject CurrentJsonObject;
    QJsonArray MyJsonObjectSelection;
    TreeStruct CurrentTreestruct;

public :
    ParisoMathObject();
    ~ParisoMathObject();
    QStringList LoadScript (QString);
    void ReadJsonFile(QString, QJsonObject &);
    int LookForIsosurfaceEquation(QString);
    int LookForParametricEquation(QString);
    int FindIsoStruct (QString);
    int FindParamStruct (QString);
    int MapIsoStruct (QString);
    void MapParamTable (QString, int);
    void MapIsoTable(QString, int);
    QString MapFunction (QString,QString);
    void MapBondary (QString,QString);
    void  MapBondary (QString, QString, int);
    void RegisterScript (QString);
    QStringList ParObject();
    QStringList IsoObject();
};
